@@ -17,6 +17,7 @@
 #include "flatpak-db.h"
 #include "flatpak-dbus.h"
 #include "flatpak-utils.h"
+#include "flatpak-dir.h"
 #include "flatpak-portal-error.h"
 #include "permission-store/permission-store-dbus.h"
 #include "xdp-fuse.h"
@@ -45,8 +46,6 @@ static GError *exit_error = NULL;
 static dev_t fuse_dev = 0;
 static GQueue get_mount_point_invocations = G_QUEUE_INIT;
 static XdpDbusDocuments *dbus_api;
-FlatpakDir *user_dir = NULL;
-FlatpakDir *system_dir = NULL;
 
 G_LOCK_DEFINE (db);
 
@@ -335,6 +334,7 @@ do_create_doc (struct stat *parent_st_buf, const char *path, gboolean reuse_exis
 static gboolean
 validate_fd_common (int fd,
                     struct stat *st_buf,
+                    mode_t st_mode,
                     char *path_buffer,
                     GError **error)
 {
@@ -353,8 +353,8 @@ validate_fd_common (int fd,
       ((fd_flags & O_NOFOLLOW) == O_NOFOLLOW) ||
       /* Must be able to fstat */
       fstat (fd, st_buf) < 0 ||
-      /* Must be a regular file */
-      (st_buf->st_mode & S_IFMT) != S_IFREG ||
+      /* Must be a regular file or directory (depending on use) */
+      (st_buf->st_mode & S_IFMT) != st_mode ||
       /* Must be able to read path from /proc/self/fd */
       /* This is an absolute and (at least at open time) symlink-expanded path */
       (symlink_size = readlink (proc_path, path_buffer, PATH_MAX)) < 0)
@@ -388,7 +388,8 @@ validate_parent_dir (const char *path,
   name = g_path_get_basename (path);
   dir_fd = open (dirname, O_CLOEXEC | O_PATH);
 
-  if (fstat (dir_fd, real_parent_st_buf) < 0 ||
+  if (dir_fd < 0 ||
+      fstat (dir_fd, real_parent_st_buf) < 0 ||
       fstatat (dir_fd, name, &real_st_buf, AT_SYMLINK_NOFOLLOW) < 0 ||
       st_buf->st_dev != real_st_buf.st_dev ||
       st_buf->st_ino != real_st_buf.st_ino)
@@ -414,7 +415,7 @@ validate_fd (int fd,
   g_autofree char *app_path = NULL;
   g_autofree char *runtime_path = NULL;
 
-  if (!validate_fd_common (fd, st_buf, path_buffer, error))
+  if (!validate_fd_common (fd, st_buf, S_IFREG, path_buffer, error))
     return FALSE;
 
   /* For apps we translate /app and /usr to the installed locations.
@@ -608,7 +609,7 @@ portal_add_named (GDBusMethodInvocation *invocation,
       return;
     }
 
-  if (!validate_fd_common (parent_fd, &parent_st_buf, parent_path_buffer, &error))
+  if (!validate_fd_common (parent_fd, &parent_st_buf, S_IFDIR, parent_path_buffer, &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
       return;
@@ -695,7 +696,7 @@ portal_lookup (GDBusMethodInvocation *invocation,
 {
   const char *filename;
   char path_buffer[PATH_MAX + 1];
-  int fd;
+  glnx_fd_close int fd = -1;
   struct stat st_buf, real_parent_st_buf;
   g_auto(GStrv) ids = NULL;
   g_autofree char *id = NULL;
@@ -1144,9 +1145,6 @@ main (int    argc,
       do_exit (4);
     }
 
-  user_dir = flatpak_dir_get_user ();
-  system_dir = flatpak_dir_get_system ();
-
   /* We want do do our custom post-mainloop exit */
   g_dbus_connection_set_exit_on_close (session_bus, FALSE);
 
